from datetime import date
from decimal import Decimal

import pytest
from dynamic_preferences.registries import global_preferences_registry
from psycopg.types.range import DateRange
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import Currency

from wbaccounting.models import BookingEntry, EntryAccountingInformation
from wbaccounting.models.entry_accounting_information import (
    VatRate,
    default_currency,
    default_email_body,
)


@pytest.mark.django_db
class TestVat:
    def test_get_rate(self, vat):
        current_rate = vat.rates.first()
        assert vat.get_rate(date(2009, 12, 31)) == Decimal("0.0")  # test exclusion
        assert vat.get_rate(date(2010, 1, 1)) == Decimal("0.08")  # test default factory data
        assert vat.get_rate(date(2050, 1, 1)) == Decimal("0.08")  # test default unbounded

        # change the end date of the existing rate and create a new rate
        new_pivot = date(2011, 1, 1)
        current_rate.timespan = DateRange(current_rate.timespan.lower, new_pivot)
        current_rate.save()
        VatRate.objects.create(vat=vat, rate=0.1, timespan=DateRange(new_pivot, date(2012, 1, 1)))
        assert vat.get_rate(date(2011, 1, 1)) == Decimal("0.1")
        assert vat.get_rate(date(2012, 1, 1)) == Decimal("0")  # excluded so default to the previous rate

        assert vat.get_rate(date(2050, 1, 1)) == Decimal("0")


@pytest.mark.django_db
class TestEntryAccountingInformation:
    def test_str(self, entry_accounting_information: EntryAccountingInformation):
        assert str(entry_accounting_information) == f"Counterparty: {entry_accounting_information.entry.computed_str}"

    @pytest.mark.parametrize(
        "method,return_value",
        [
            ("get_endpoint_basename", "wbaccounting:entryaccountinginformation"),
            ("get_representation_value_key", "id"),
            ("get_representation_label_key", "{{entry_repr}}"),
            ("get_representation_endpoint", "wbaccounting:entryaccountinginformationrepresentation-list"),
        ],
    )
    def test_wbmodel_methods(self, method: str, return_value: str):
        assert getattr(EntryAccountingInformation, method)() == return_value

    @pytest.mark.parametrize("user__is_superuser", [False])
    def test_filter_for_user_no_superuser(self, entry_accounting_information: EntryAccountingInformation, user: User):
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information not in entry_accounting_information_list

    @pytest.mark.parametrize("user__is_superuser", [True])
    def test_filter_for_user_superuser(self, entry_accounting_information: EntryAccountingInformation, user: User):
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information in entry_accounting_information_list

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [False])
    def test_filter_for_user_public_counterparty(
        self, entry_accounting_information: EntryAccountingInformation, user: User
    ):
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information in entry_accounting_information_list

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty(
        self, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information not in entry_accounting_information_list

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty_with_exempt(
        self, booking_entry: BookingEntry, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        entry_accounting_information.exempt_users.add(user)
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information in entry_accounting_information_list

    def test_default_email_body(self):
        assert default_email_body() == ""

    def test_custom_default_email_body(self):
        global_preferences_registry.manager()["wbaccounting__invoice_email_body"] = "Custom Body"
        assert default_email_body() == "Custom Body"

    def test_default_currency(self):
        assert default_currency() is None

    def test_custom_default_currency(self, currency: Currency):
        global_preferences_registry.manager()["wbaccounting__default_entry_account_information_currency_key"] = (
            currency.key
        )
        assert default_currency().pk is currency.pk



from contextlib import suppress
from datetime import date
from decimal import Decimal

from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateRangeField, RangeOperators
from django.db import models
from django.db.models import Q, QuerySet, UniqueConstraint
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.module_loading import import_string
from dynamic_preferences.registries import global_preferences_registry as gpr
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import Currency
from wbcore.signals import pre_merge
from wbcore.utils.models import PrimaryMixin

from wbaccounting.models.model_tasks import generate_booking_entries_as_task


def default_email_body() -> str:
    with suppress(Exception):
        return gpr.manager()["wbaccounting__invoice_email_body"]

    return ""


def default_currency() -> Currency | None:
    with suppress(Exception):
        return Currency.objects.get(key=gpr.manager()["wbaccounting__default_entry_account_information_currency_key"])


class EntryAccountingInformationDefaultQuerySet(QuerySet):
    def filter_for_user(self, user: User) -> QuerySet:
        """
        Filters entry accounting information based on if the current user is allowed to see it.

        Args:
            user (User): The user for whom entry accounting information need to be filtered.

        Returns:
            QuerySet: A filtered queryset.
        """

        if user.is_superuser or user.has_perm("wbaccounting.administrate_invoice"):
            return self

        if not user.has_perm("wbaccounting.view_entryaccountinginformation"):
            return self.none()

        return self.filter(Q(counterparty_is_private=False) | Q(exempt_users=user))


class Vat(PrimaryMixin, models.Model):
    name = models.CharField(max_length=255)

    def get_rate(self, date: date) -> Decimal:
        try:
            return self.rates.get(timespan__contains=date).rate
        except VatRate.DoesNotExist:
            return Decimal("0.0")

    def __str__(self) -> str:
        return self.name

    class Meta:
        verbose_name = "Vat Group"
        verbose_name_plural = "Vat Groups"
        constraints = [
            UniqueConstraint(
                name="unique_primary_vat_group",
                fields=["primary"],
            )
        ]

    @classmethod
    def get_representation_endpoint(cls):
        return "wbaccounting:vatrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"


class VatRate(models.Model):
    vat = models.ForeignKey("wbaccounting.Vat", related_name="rates", on_delete=models.CASCADE)
    timespan = DateRangeField(verbose_name="Timespan")
    rate = models.DecimalField(decimal_places=4, max_digits=5)

    class Meta:
        verbose_name = "Vat Rate"
        verbose_name_plural = "Vat Rates"
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_vat_rates",
                expressions=[
                    ("timespan", RangeOperators.OVERLAPS),
                    ("vat", RangeOperators.EQUAL),
                ],
            ),
        ]

    def __str__(self) -> str:
        return f"{self.vat}: {self.rate}"


class EntryAccountingInformation(models.Model):
    # Link to Entry
    entry = models.OneToOneField(
        "directory.Entry",
        on_delete=models.CASCADE,
        related_name="entry_accounting_information",
        verbose_name="Linked Counterparty",
    )

    # Tax Information
    tax_id = models.CharField(max_length=512, blank=True, null=True, verbose_name="Tax ID")
    vat = models.ForeignKey(
        "wbaccounting.Vat", related_name="vat_recipients", on_delete=models.PROTECT, verbose_name="VAT"
    )

    # Invoice Information
    default_currency = models.ForeignKey(
        "currency.Currency",
        related_name="entry_accounting_informations",
        default=default_currency,
        on_delete=models.PROTECT,
        verbose_name="Default Currency",
        blank=True,
        null=True,
    )

    email_to = models.ManyToManyField(
        "directory.EmailContact", related_name="entry_accounting_informations_to", blank=True, verbose_name="To"
    )

    email_cc = models.ManyToManyField(
        "directory.EmailContact", related_name="entry_accounting_informations_cc", blank=True, verbose_name="CC"
    )
    email_bcc = models.ManyToManyField(
        "directory.EmailContact", related_name="entry_accounting_informations_bcc", blank=True, verbose_name="BCC"
    )

    email_subject = models.CharField(default="{{invoice.title}}", max_length=1024, verbose_name="Subject")
    email_body = models.TextField(default=default_email_body, verbose_name="Body")
    send_mail = models.BooleanField(default=True, verbose_name="Send Mail")

    counterparty_is_private = models.BooleanField(
        default=False,
        verbose_name="Counterparty Is Private",
        help_text="Hides all of the counterparty's invoices from non-eligible users",
    )
    exempt_users = models.ManyToManyField(
        to="authentication.User",
        verbose_name="Exempt Users",
        help_text="Exclusion list of users who are able to see private invoices for the counterparty",
        related_name="private_accounting_information",
        blank=True,
    )
    booking_entry_generator = models.CharField(max_length=256, null=True, blank=True)
    default_invoice_type = models.ForeignKey(
        to="wbaccounting.InvoiceType",
        related_name="booking_entries",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Default Invoice Type",
        help_text="When invoicinging outstanding booking entries, this invoice type will be assigned to the corresponding invoice",
    )

    external_invoice_users = models.ManyToManyField(
        to="authentication.User",
        verbose_name="External User",
        help_text="External users who are able to see the invoices generated for this counterparty",
        related_name="external_accounting_information",
        blank=True,
    )

    def get_booking_entry_generator(self):
        with suppress(ImportError):
            return import_string(self.booking_entry_generator or "")

    objects = EntryAccountingInformationDefaultQuerySet.as_manager()

    def save(self, *args, **kwargs):
        if not self.vat:
            self.vat = Vat.objects.get_or_create(primary=True, defaults={"name": "Default VAT group"})[0]
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Counterparty"
        verbose_name_plural = "Counterparties"

    def __str__(self):
        return f"Counterparty: {self.entry.computed_str}"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbaccounting:entryaccountinginformationrepresentation-list"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbaccounting:entryaccountinginformation"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{entry_repr}}"

    def generate_booking_entries(self, from_date: date, to_date: date, booking_date: date):
        generate_booking_entries_as_task.delay(  # type: ignore
            self.booking_entry_generator or "", from_date, to_date, self.entry.id, booking_date
        )


@receiver(post_save, sender=EntryAccountingInformation)
def post_save_entry(sender, instance, **kwargs):
    """If the EAI does not have any email_to, then we add the entries primary email address to it (if it exists)"""
    if not instance.email_to.exists() and (email := instance.entry.primary_email_contact()):
        instance.email_to.add(email)


@receiver(pre_merge, sender="directory.Entry")
def pre_merge_entry_entryaccountinginformation(sender: models.Model, merged_object, main_object, **kwargs):
    with suppress(EntryAccountingInformation.DoesNotExist):
        rel = EntryAccountingInformation.objects.get(entry=merged_object)
        if not EntryAccountingInformation.objects.filter(entry=main_object).exists():
            rel.entry = main_object
            rel.save()
        else:
            rel.delete()



from django.contrib import admin

from wbaccounting.models import EntryAccountingInformation, Vat, VatRate


class VarRateInline(admin.TabularInline):
    model = VatRate
    fk_name = "vat"
    fields = ("rate", "timespan")


@admin.register(Vat)
class VatModelAdmin(admin.ModelAdmin):
    list_display = ("name",)
    search_fields = ["name"]

    inlines = [VarRateInline]


@admin.register(EntryAccountingInformation)
class EntryAccountingInformationModelAdmin(admin.ModelAdmin):
    list_display = ("entry", "send_mail", "counterparty_is_private")
    autocomplete_fields = ["default_currency", "entry", "email_to", "email_cc", "email_bcc", "exempt_users"]

    search_fields = ["entry__computed_str"]



from wbcore.menus import ItemPermission, MenuItem

ENTRYACCOUNTINGINFORMATION_MENUITEM = MenuItem(
    label="Counterparties",
    endpoint="wbaccounting:entryaccountinginformation-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal,
        permissions=["wbaccounting.view_entryaccountinginformation"],
    ),
)
